#include "PosHistory.hpp"
#include "PosTracker.hpp"
#include "filter.hpp"
#include "robot.hpp"
#include "utils.hpp"

#include <SFML/Graphics/PrimitiveType.hpp>
#include <SFML/Graphics/RectangleShape.hpp>
#include <SFML/Graphics/RenderWindow.hpp>
#include <SFML/Graphics/Transform.hpp>
#include <SFML/Graphics/Transformable.hpp>
#include <SFML/Graphics/Vertex.hpp>
#include <SFML/Graphics/VertexArray.hpp>
#include <SFML/System/Vector2.hpp>
#include <SFML/Window/Event.hpp>
#include <SFML/Window/Keyboard.hpp>
#include <csignal>
#include <iostream>
#include <sstream>

using sf::Vector2f;

using namespace std::chrono_literals;

constexpr auto THRESHOLD = 10;

void controller_follow_border_step(Robot& robot, float speed_factor) {
    auto line_value = DBG(int(robot.getLineTracksValue(0)));
    auto detects_line = line_value > THRESHOLD;
    if (detects_line) robot.setMotorsPower(0.2 * speed_factor, 0.1 * speed_factor);
    else robot.setMotorsPower(0.1 * speed_factor, 0.2 * speed_factor);
}

void controller_follow_border(Robot &robot, float speed_factor) {
    while (true) {
        auto events = {EventType::LINE_TRACKS_IS_DETECTED,EventType::SWITCHS_IS_DETECTED};
        robot.waitChanged(events, 1s);
        controller_follow_border_step(robot, speed_factor);
    }
}

// note : Ne marche pas car remplace le stack du code interrompu.
static bool do_stop = false;
void stop_handler(int) {
    do_stop = true;
}

void event(sf::RenderWindow & window, Robot& robot) {
    bool do_stop_ = do_stop;
    sf::Event event;
    while (window.pollEvent(event)) {
        if (event.type == sf::Event::Resized) {
            sf::FloatRect visibleArea(0.f, 0.f, event.size.width, event.size.height);
            window.setView(sf::View(visibleArea));
        }
        if (event.type == sf::Event::Closed) do_stop_ = true;
        if (event.type == sf::Event::KeyPressed && event.key.code == sf::Keyboard::Escape) do_stop_ = true;
    }
    if (do_stop_) {
        robot.setMotorPower(Robot::MotorIndex::LEFT, 0);
        robot.setMotorPower(Robot::MotorIndex::RIGHT, 0);
        window.close();
    }
}


constexpr float ROBOT_SIZE = 20.0;
constexpr float WHEEL_DIAMETER = 6.5;
constexpr float LATICE_COUNT = 20;


class ControlWindow {
    private:
        Robot& _robot;
        sf::RenderWindow _window;
        sf::RectangleShape _robotShape;
        PosTracker _tracker;
        PosHistory _history;
        MotorJitterFilter _filter;
        float _speed_factor;

    public:
        ControlWindow(Robot& robot, Vector2f robot_size, float wheel_diameter, float latice_count, int history_count, float speed_factor, float filter_delay) :
            _robot(robot),
            _window(sf::RenderWindow(
            sf::VideoMode(800u, 600u),
            "robot-command",
            sf::Style::Titlebar | sf::Style::Close
            )),
            _robotShape(sf::RectangleShape(robot_size)),
            _tracker(PosTracker(Vector2f(90, 90), 0., robot_size.y, wheel_diameter, latice_count)),
            _history(PosHistory(history_count)),
            _speed_factor(speed_factor),
            _filter(filter_delay)
        {
            _window.setVerticalSyncEnabled(true);
            _robotShape.setOrigin(sf::Vector2f(robot_size.x / 2, robot_size.y / 2));
            _robotShape.setFillColor(sf::Color::Blue);
            _tracker.update(robot.getEncoderWheelValue(1), robot.getEncoderWheelValue(0));
            _tracker.reset(Vector2f(90, 90), 0.);
        }

        void run() {
            while (_window.isOpen()) {
                update();
                controller_follow_border_step(_robot, _speed_factor);
            }
        }

    private:
        void update() {
            _tracker.update(_robot.getEncoderWheelValue(1), _robot.getEncoderWheelValue(0));

            _robotShape.setPosition(_tracker.getPosition());
            _robotShape.setRotation(_tracker.getRotation());
            _history.append(_tracker.getPosition());

            event(_window, _robot);
            _window.clear();
            _window.draw(_robotShape);
            _window.draw(_history.vertices());
            _window.display();
        }
};


int main(int argc, char ** argv)
{
    signal(SIGKILL, stop_handler);
    std::string hostIpAddress("127.0.0.1");
    uint16_t tcpPort = 6543;

    if (argc == 3)
    {
        hostIpAddress = std::string(argv[1]);
        std::istringstream iss(argv[2]);
        iss.exceptions(std::istringstream::failbit);
        iss >> tcpPort;
    }

    std::cout << "Try to connect to " << hostIpAddress << ":" << tcpPort << " ..." << std::endl;
    Robot robot(hostIpAddress, tcpPort);
    robot.waitReady();

    try {
        // #define STEP_1
        #ifdef STEP_1
        controller_follow_border(robot);
        #else
        ControlWindow(robot, Vector2f(ROBOT_SIZE, ROBOT_SIZE), WHEEL_DIAMETER, LATICE_COUNT, 4000, 0.78, 1).run();
        #endif
    } catch (std::exception &e) {
      std::cerr << e.what() << std::endl;
    }

    return 0;
}
